/*
 * An oracle for finding solutions of a^2 + b^2 = -1.
 * This will never invoke a norm equation solver and only performs
 * cheap computations... it will thus in general fail.
 * There are two ingredients: (i) Real-time solutions if they are
 * theoretically known and (ii) a database of precomputed solutions.
 * The latter are stored as follows:
 *
 * Even type: <p, +1, a, b>, where a and b are the coefficients of
 * polynomials in a primitive p-th root of unity; ord(2 mod p) is even.
 * Note that we could also construct these entries on the fly.
 *
 * Odd type:  <p, -1, a, b>, where a = [a0, a1], b = [b0, b1] corr to
 * a = a0 + sqrt(2) * a1; the ai and bi are again coeff of polynomials;
 * ord(2 mod p) is odd.
 */
 
import "misc.m":  MyIsFunctionField,
    MinimalPolynomialSemisimpleMatrix, PandPPrimePart, CyclotomicExtension;
 
SO2_solutions := [*
<7, -1, [[-1,-1,-1,0,-1,0], [1/2,1/2,1/2,0,1/2,0]], [[-1,0,0,0,0,0],[1/2,-1/2,-1/2,0,-1/2,0]]>
    *];

/*
 * The smallest missing entries are 23(176), 31(240) 47(368), 71(560).
 * The numbers in brackets denote the smallest corresponding dimensions for
 * which irreducibility testing then fails over the rationals.
 */

/*
 * SO2_retrieve
 *
 * If possible, either fetch a solution of a^2 + b^2 = -1 from the library
 * or construct it on the fly.
 *
 * If allow_sqrt2 = false, then only consider solutions in the p-th
 * (p a prime divisor of m) cyclotomic field; otherwise sqrt(2) may be
 * adjoined.
 *
 * For the construction, we use the formula given in
 *     Huppert: ``Character theory of finite groups'', Example 38.13d.
 * Explicitly, let e = zeta_p where r = ord(2 mod p) is even.
 * Let d = e^(-2^(r-1)). Then
 *     -1 = Norm(d * prod_(t=0)^(r-1) (d^(2^t) + i)).
 */

function SO2_retrieve(m, allow_sqrt2)
    S := PrimeDivisors(m);
    for sol in SO2_solutions do
        if (sol[1] in S) and ((sol[2] eq 1) or allow_sqrt2) then
            // Q: Why does `Explode' fail here?
            return true, sol[1], sol[2], sol[3], sol[4];
        end if;
    end for;
    for p in S do
        ord := Order(Integers(p)!2);
        if IsEven(ord) then
                K<e> := CyclotomicField(p);
                r := ExactQuotient(ord, 2);
    
               /*
                * Represent F = K(i) as tuples; this is faster than
                * constructing F as a field via ext<...>.
                */
                d := e^(-2^(r-1));
                mul := func< x, y |
                    <x[1] * y[1] - x[2]*y[2], x[1] * y[2] + x[2] * y[1]>
                    >;
                x := <d,0>;
                for t in [0..r-1] do
                    x := mul(x, <d^(2^t),1>);
                end for;
                return true, p, +1, Eltseq(x[1]), Eltseq(x[2]);
        end if;
    end for;
    return false, _, _, _, _;
end function;

/*
 * IsIrreducibleANC
 *
 * Irreducibility testing of non-abelian ANC groups with homogeneous
 * cyclic maximal subgroup and known structure.
 * The input will not be given as a `GrpMat'; MatrixGroup<..> is slow!
 *
 * INPUT:  G1^2 = theta, G2 generates a homogeneous cyclic subgroup of
 *         index 2; the `order' of <G1,G2>, `mu' and `DecideOnly'
 *         as usual
 * OUTPUT: see `IsIrreducibleRelative'
 */
 
function IsIrreducibleANC(G1, G2, order, theta 
           :  mu := false, DecideOnly := false, NeqnThreshold := Infinity())
    vprint Irrednil: "IsIrreducibleANC";               
    d := Nrows(G1);
    vprintf Irrednil: "\t order = %o, theta = %o, degree = %o\n",
        order, theta, d;

    apply_mu := func<x | mu cmpeq false select x else x * mu>;
        
   /*
    * Get the `FldFunRat/FldNum/FldRat' decomposition.
    */        
    K := BaseRing(G1);

   /*
    * Since this is not a user-level function, we assume that we only
    * sensible input.
    */

    k := MyIsFunctionField(K) select BaseRing(K) else K;
    V := VectorSpace(K, d);
    A := EndomorphismAlgebra(V);

   /*
    * By working with reps of free groups, we avoid Magma's
    * MatrixGroup<...>  constructor. While this saves some time, setting
    * up the GModule-structure is still the major bottleneck here!
    */
    
    VG := GModule(FreeGroup(2), [G1,G2]);
    
   /*
    * Get dimension of the irreducible K[G]-module.
    */
    vprintf Irrednil: "\t getting degree of maximal subfield of K[G]: ";
    phi := Type(k) eq FldRat
             select EulerPhi(ExactQuotient(order,2))
             else Degree(MinimalPolynomialSemisimpleMatrix(G2));
    vprintf Irrednil: "%o\n", phi;

   /*
    * If the cyclic maximal subgroup is irreducible, then we are done.
    * Note that we do not actually need to spin up any K[A]-submodule to
    * detect this case.
    */           
    if d eq phi then
        vprint Irrednil: "\t the cyclic maximal subgroup is irreducible";
        return true, _;
    end if;

    
   /*
    * Check if for U = V.1 * K[G2], we have V = U (+) U * G1 as required
    * by §6.2. In particular, we then have d = 2 * phi.
    */
    vprintf Irrednil:
        "\t attempting to construct system of imprimitivity... ";
    if Dimension(sub<VG|V.1>) lt Dimension(V) then
        vprint Irrednil: "failed"; // NOTE: failure is good!
        if DecideOnly then
            return false, _;
        else
            return false, apply_mu(V.1);
        end if;
    end if;
    vprint Irrednil: "success";

    case theta:
    when 1:
       /*
        * Dihedral or semidihedral type. 
        */
        if DecideOnly then
            return false, _;
        else
            // Q: What's the cheapest way to get just ONE non-zero
            //    element from the kernel of a matrix?
            //    This just doesn't seem right!
            return false, apply_mu(Kernel(A!G1 + 1).1);
        end if;
        
    when -1:
       /*
        * Generalised quaternion type of order 2^n * m.
        */
	twon, m := PandPPrimePart(order, 2);
	n := Valuation(twon,2);
        _, u, v := Xgcd(2^(n-1), m);
        	
       /*
        * Find the dimension of the faithful irreducible module.
        * In other words, decide whether a^2 + b^2 = -1 can be solved
        * in the associated cyclotomic extension of `k'.
        * If this is the case, then `e = 1', otherwise `e = 2'.
        */
        vprintf Irrednil: "\t getting dimension of the irreducible module: ";
        if Type(k) eq FldRat then
            e := (m gt 1) and ((n ge 4) or IsEven(Order(Integers(m)!2)))
                 select 1 else 2;
        else
            // Lazy evaluation: Cheap tests first!
            e :=
              ((m gt 1) or IsEmpty(RealPlaces(k)))
                 and
             (((m gt 1) and IsEven(Order(Integers(m)!2)))
                                or forall{<> : d in Decomposition(k,2)
                                             | IsEven(LocalDegree(v))
                                             where v is d[1]
                                           })
             select 1 else 2;                                           
        end if;
        vprint Irrednil: e * phi;
        
	if d eq e * phi then
	    return true, _; // this can only happen for e = 2!
        elif DecideOnly then
            return false, _;
        end if;
        
       /*
        * Now we actually need to solve a^2 + b^2 = -1.
        */
        vprint Irrednil: "\t we need to solve a^2 + b^2 = -1";

       /*
        * (2,2')-decomposition G2 = x * y.
        */
        g := A ! G1;
	x := A ! G2^(v * m);
	y := A ! G2^(u * 2^(n-1));
        z := A ! G2;

       /*
        * First attempt: Ask the oracle.
        */
        repeat
            vprintf Irrednil: "\t does the oracle know a solution? ";
            // the field K[A] contains sqrt(2) if n >= 4
            // TODO: also include the case that sqrt(2) is in K 
            r, p, resord, seqa, seqb := SO2_retrieve(m, n ge 4);
            if r then
                vprintf Irrednil: "yea (prime = %o)\n", p;
            else
                vprintf Irrednil: "nay\n";
                break;
            end if;

            vprintf Irrednil: "\t decoding the oracle's answer... ";
            if n ge 4 then
                sqrti := x^(2^(n-4));
                i := sqrti^2;
            else
                i := x;
            end if;
            zeta := y^(ExactQuotient(m, p));
            
            // construct solution
            case resord:
            when +1:
                a := Evaluate(Polynomial(seqa), zeta);
                b := Evaluate(Polynomial(seqb), zeta);
            when -1:
                // NOTE: only happens for n >= 4; sqrti is defined
                sqrt2 := sqrti + sqrti^(-1);
                a := Evaluate(Polynomial(seqa[1]),zeta)
                    + Evaluate(Polynomial(seqa[2]),zeta) * sqrt2;
                b := Evaluate(Polynomial(seqb[1]),zeta)
                    + Evaluate(Polynomial(seqb[2]),zeta) * sqrt2;
            end case;
            vprint Irrednil: "done";
            sol := a + b * i;
        until true;
        
       /* 
        * Second attempt: Use a norm equation solver or abort computation
        *                 if the degree is too large.
        */
        repeat
            if assigned sol then break; end if;
            // top degree = phi * AbsoluteDegree(k)
            if AbsoluteDegree(k) * phi gt NeqnThreshold then 
                vprintf Irrednil: "\
\t threshold for norm equation solver exceeded (top degree = %o) 

\t To force the computation of a solution, set `NeqnThreshold' to a
\t higher value and repeat the computation. This may take a long time...
\n", AbsoluteDegree(k) * phi;
                return false, _;
            end if;
            
            C<zeta> := CyclotomicExtension(k, 2^(n-1) * m);
            zeta_2  := zeta^(v * m);
            zeta_m  := zeta^(u * 2^(n-1));
            F := sub<C | zeta_2 + zeta_2^(-1), zeta_m>;
            L := ext<F | Polynomial([1,0,1])>;
            
            vprintf Irrednil:
                "\t solving norm equation (degree = %o)... ",
                AbsoluteDegree(L);
            r, s := NormEquation(L, F!(-1));
            vprint Irrednil: "done";
            error if not r,
                "norm equation does not have a solution; \
this should never happen";

           /*
            * We have: a + b * i in L.
            * Get a,b in F, coerce into C so that they are polynomials
            * in zeta, hence in z.
            */
            seqa, seqb := Explode([C!x:x in Eltseq(s[1])]);
            a := Evaluate(Polynomial(Eltseq(seqa)),z);
            b := Evaluate(Polynomial(Eltseq(seqb)),z);
            sol := a + (x^(2^(n-3))) * b;
        until true;
        vprintf Irrednil: "\t recovering generator of submodule... ";
        x := apply_mu(Kernel(g - sol).1);
        vprint Irrednil: "done";
        return false, x;
    end case;
end function;
