import "anc.m": SO2_retrieve;
import "misc.m": MyIsFunctionField, CanHandleField, ResidueClassOrder;

function RationaliseMatGrp(G)
    d := Degree(G);
    H := GL(d,Rationals());
    return MatrixGroup<d, Rationals() | {H!g : g in Generators(G)}>;
end function;

// Rewrite a matrix group defined over a number field over the rationals.
function RewriteOverRationals(G)
    if BaseRing(G) cmpeq Rationals() then
	return G;
    elif BaseRing(G) cmpeq Integers() then
	return RationaliseMatGrp(G);
    end if;

    d := Degree(G);
    K := AbsoluteField(BaseRing(G));

    B := Basis(K);
    V, f := KSpace(K, Rationals());
    // isomorphism from K into End(V)
    iota := func<a |
	Matrix([(b * a)@f : b in B])
	>;

    sigma := func< g |
	BlockMatrix( [[ g[i,j]@iota : j in [1..d]] : i in [1..d]])
	>;
    G := MatrixGroup< d * Degree(K), Rationals() |
	[g @ sigma : g in Generators(G) ]
	>;
    return G;
end function;

// Apply a field homomorphism coefficientwise to a matrix group
function ApplyFieldHomMatGrp(G, f)
    return G where G is
        MatrixGroup<Degree(G),Codomain(f) |
        [ Eltseq(G.i)@f : i in [1..Ngens(G)]]>;
end function;

// generate a `random' invertible rational n*n-matrix with coefficients
// of absolute values at most `thres'

// frac = true: rational coefficients
function RandomMat(args, ... : thres := false, frac := false)
    n := args[1];
    if #args eq 1 then
        K := Rationals();
    elif #args eq 2 then
        K := args[2];
    else
        error "bad arguments";
    end if;
    
    error if not CanHandleField(K),
        "I don't know what to do with this field";
    
    // default values for `thres' differ depending on whether `frac' or not
    if thres cmpeq false then
        thres := 2;
    end if;
    
    R := Integers(K); // either a polynomial ring or the maximal order
    
    if frac then
        if MyIsFunctionField(K) then
            rand := func<| Random(K,thres)/
                ((w eq 0) select 1 else w) where w is Random(K,thres) >;
        else
            rand := func<|Random(K,thres)>;
        end if;
    else
        if Type(R) eq RngInt then
            rand := func<|Random(-thres+1,thres-1)>;
        elif Type(R) in {RngUPol,RngMPol} then
            rand := func<|Random(K,thres)>;
        else
            // maximal order of a number field
            rand := func<| R![Random(-thres+1, thres-1) 
                 : i in [1..#Basis(R)]]>;
        end if;
    end if;

    /*if MyIsNumberField(K) then
        k := MyAbsoluteField(K);
    else
        k := MyAbsoluteField(BaseRing(K));
    end if;*/

    //A := [ [] : t in [1..n]];
    repeat
        A := [ [rand() : i in [1..n]] : j in [1..n]];
        //for i, j in [1..n] do
        //A[i,j] := k! [ Random(-thres,thres)/Random([-thres..-1] cat [1..thres]) : r in [1..Degree(k)]];
	//end for;
	A := Matrix(K, A);
    until Determinant(A) ne 0;
    return GL(n,K) ! A;
end function;

function IntegralRandomify(x, N)
    d := Nrows(x);
    for counter in [1..N] do
        if Random(1) eq 0 then
            SwapRows(~x, Random(1,d),Random(1,d));
        else
            SwapColumns(~x, Random(1,d), Random(1,d));
        end if;

        if Random(1) eq 0 then
            i := Random(1,d);
            j := Random([1..i-1] cat [i+1..d]);
            AddRow(~x, Random(-2,2), i, j);
        else
            i := Random(1,d);
            j := Random([1..i-1] cat [i+1..d]);
            AddColumn(~x, Random(-2,2), i, j);
        end if;
    end for;
    return x;
end function;

// given an abstract group G and a linear rep f (just a `map' in Magma)
// of a subgroup H of G, create the image of the induced rep of G
function InducedRepMatrixGroup(G, f)
    local x;
    
    H := Domain(f);
    T, tau := Transversal(G,H);

    d := Nrows(Id(H) @ f);
    R := BaseRing(Id(H) @ f);
    V := VectorSpace(R, d);

    act := [];
    for g in Generators(G) do
	M := [];
	for t in T, v in Basis(V) do
	    w := [V!0 : t in T];
	    x := t * g;
	    w[Index(T,x@tau)] := v * (H!(x/x@tau))@f;
	    Append(~M, &cat[Eltseq(x) : x in w]);
	end for;
	Append(~act, Matrix(M));
    end for;
    K := MatrixGroup< d*#T, R | act>;
    if #G eq 1 then
	return K;
    else
	return K;
    end if;
end function;

function SylowRatGL(d,p : Fast := true)
    //requirege d, 1;
    //require IsPrime(p): "second argument is not a prime";
    
    Q := Rationals();
    
    if d eq 1 then return sub<GL(1,Q)|>; end if;
    
    r := Valuation(d div (p-1),p) + 1;
    if d eq (p-1) * p^(r-1) then
        G := MatrixGroup<p-1,Q|
               CompanionMatrix(CyclotomicPolynomial(p))>;
        if Fast then
            // fast mode: G wr (Cp wr Cp wr ...)               
            P := CyclicGroup(1);
            for i in [1..r-1] do
                P := WreathProduct(P,CyclicGroup(p));
            end for;
            return WreathProduct(G,P);
        else
            // slow mode: ((G wr Cp) wr Cp) wr ...
            for i in [1..r-1] do
                G := WreathProduct(G,CyclicGroup(p));
            end for;
            return G;
        end if;
    end if;

    a := Intseq(d div (p-1), p);
    if IsEmpty(a) then return sub<GL(d,Q)|>; end if;

    G := false;
    for i in [1..#a] do
        if a[i] eq 0 then continue; end if;
        P := DirectProduct( [ R : j in [1..a[i]]] )
               where R is $$((p-1) * p^(i-1), p : Fast := Fast);
        G := (Type(G) eq BoolElt) select P else DirectProduct(G,P);           
    end for;

    return a0 eq 0 select G else DirectProduct(G,sub<GL(a0,Q)|>)
        where a0 is d mod (p-1);
end function;

// Kronecker product of two matrix groups
function KronProd(G, H)
    return G where G:= MatrixGroup< Degree(G) * Degree(H), BaseRing(G) |
	[ KroneckerProduct(g, Id(H)) : g in Generators(G) ]
        cat
	[ KroneckerProduct(Id(G), h) : h in Generators(H) ]
	>;
end function;

// blow up Generators(G) to size N, then run K iterations of product
// replacement
function RandomifyMatGrp(G, N, K)
    q := N div Ngens(G);
    r := N mod Ngens(G);
    
    gens := [G.i : i in [1..Ngens(G)]];
    S := &cat[gens : i in [1..q]] cat [Id(G) : i in [1..r]];

    for count in [1..K] do
        i := Random([1..N]);
        j := Random([1..i-1] cat [i+1..N]);
        if Random(1) eq 1 then
            S[i] := S[i] * S[j];
        else
            S[i] := S[j] * S[i];
        end if;
    end for;
    return G where G is MatrixGroup<Degree(G), BaseRing(G)| S>;
end function;

// L = list of ring generators
function RegularQuaternionAlgebra(L, a, b)
    if Type(L[1]) notin {AlgMatElt, GrpMatElt} then
        R := Parent(L[1]);
        L := [ Matrix(R,1,1,[x]) : x in L];
    end if;

    I := L[1]^0;
    O := 0 * I;

    return [DiagonalJoin([x,x,x,x]) : x in L]
    cat [
    BlockMatrix(4,4, [  O, I,  O,  O,
                      a*I, O,  O,  O,
                        O, O,  O,  -I,
                        O, O,-a*I,  O]),

    BlockMatrix(4,4, [  O, O,  I,  O,
                        O, O,  O,  I,
                      b*I, O,  O,  O,
                        O,b*I, O,  O])
    ];
end function;

/*
 * SO2_solve_in_field
 *
 * Let K be the m-th cyclotomic field. Assuming the library permits it,
 * if -1 = a^2 + b^2 in F = K or F = K(sqrt 2); return F and a
 * solution a, b.
 *
 * This function is not needed for irreducibility testing but it can be
 * used to construct examples or irreducible ANC groups.
 */

 
function SO2_solve_in_field(m)
    if IsDivisibleBy(m,4) or IsDivisibleBy(2, m) then
	K := CyclotomicField(Lcm(4,m));
	F := CyclotomicField(4);
	return K, K!F.1, K!0;
    end if;
    S := PrimeDivisors(m);

    r, p, eps, a, b := SO2_retrieve(m,true);
    if not r then
        return false, _, _;
    end if;
    case eps:
    when +1:
	m0 := m;
	K0 := CyclotomicField(m0);
	m := IsEven(m0) select (m0 div 2) else m0;
	K := CyclotomicField(m);
	z := K.1^(m div p);
	F := sub< K | z>;
	return K0, K0!K!F!a, K0!K!F!b;
    when -1:
	m0 := m;
	K0 := CyclotomicField(m0);
	L0<sqrt2> := ext<K0 | Polynomial([-2,0,1])>;

	m := IsEven(m0) select (m0 div 2) else m0;
	K := CyclotomicField(m);
	z := K!CyclotomicField(p).1;
	F := sub< K | z>;
	return L0, (K0!K!F!a[1]) + (K0!K!F!a[2]) * sqrt2,
	    (K0!K!F!b[1]) + (K0!K!F!b[2]) * sqrt2;
    end case;
end function;

/*
 * construct irreducible versions of type +1 ANC groups over
 * minimal splitting fields
 */
IrrDihANC := function(arg, ...)
    if #arg eq 2 then
	n := arg[1]; m := arg[2];
    elif #arg eq 1 then
	n := arg[1]; m := 1;
    else
	error "";
    end if;

    K<z> := CyclotomicField(2^(n-1) * m);
    x := z^m;
    w := z^(2^(n-1));

    L := sub<K | x + x^(-1), w>;
    return G where G := MatrixGroup<2,L|[
	Matrix(K, [[0,1],[-1,x+x^(-1)]]),
	Matrix(K, [[0,w],[w,0]])
	]>;
end function;

// after restricting to Q, this group is reducible iff
// -1 = a^2 + b^2 in the field Z from the paper
RarelyIrrQuatANC := function(arg, ...)
    if #arg eq 2 then
	n := arg[1]; m := arg[2];
    elif #arg eq 1 then
	n := arg[1]; m := 1;
    else
	error "";
    end if;

    K<z> := CyclotomicField(2^(n-1) * m);
    x := z^m;
    w := z^(2^(n-1));

    return G where G := MatrixGroup<2,K |
	Matrix([[x,0],[0,x^(-1)]]),
	Matrix([[0,w],[-w,0]])>;
end function;

// construct the irreducible ANC group of order 2^n*m and generalised
// quaternion type... if we can do this!
// the choice of basis is currently pretty bad: matrices have huge entries
function IrrQuatANC(arg, ...)
    if #arg eq 2 then
	n := arg[1]; m := arg[2];
    elif #arg eq 1 then
	n := arg[1];
	m := 1;
    else
	error "";
    end if;

    if (m eq 1) or (n eq 3 and IsOdd(ResidueClassOrder(2,m))) then
	return RarelyIrrQuatANC(n,m);
    end if;
    
    K, a, b := SO2_solve_in_field(m);
    phi := Factorisation( PolynomialRing(K) !
	CyclotomicPolynomial(2^(n-1)) )[1,1];
    L<zeta> := ext<K | phi>;
    F := sub<L | zeta+zeta^(-1)>;
    L<i> := ext<F |Polynomial([1,0,1])>;
    zeta := Rep(Roots(
	PolynomialRing(L) ! CyclotomicPolynomial(2^(n-1)) ))[1];
    omega := Rep(Roots(
	PolynomialRing(L) ! CyclotomicPolynomial(m)))[1];
    a := F!a; b := F!b;
    V, f := VectorSpace(L, F);
    A := Matrix([1@f,i@f]);
    // adjust basis so that 1 -> (1,0), i -> (0,1)
    f := map< L->V | x :-> x@f * A^(-1)>;
    x := omega * Matrix([zeta@f, (i*zeta)@f]);

    g := Matrix([[a,b],[b,-a]]);
    return G where G is MatrixGroup<2, F | x,g>;
end function;

function conjmat(d,K)
    if Type(K) eq FldRat then
        return Id(GL(d,K));
    end if;

    x := K.1;
    
    // the `3' in the denominator makes sure that the first choice of an
    // admissable prime does NOT always succeed
    A := Matrix(2,2,[1+x,1/(3*x),0,1-x]);
    O := Matrix(2,2,[K|0,0,0,0]);
    a := DiagonalJoin(A,A);
    o := DiagonalJoin(O,O);
    
    t8 := BlockMatrix(4,4,
            [A,O,O,O,
             O,O,O,A,
             A,O,A,O,
             O,A^(-1),O,O]);
    t12 := BlockMatrix(3,3,
            [
            o,o,a,
            a,o,o,
            o,a,o
            ]);
    t16 := BlockMatrix(4,4,
              [o,a,o,a,
               o,o,a,o,
               o,a^(-1),o,a,
               a,o,o,o]);
    t20 := BlockMatrix(5,5,
             [a^(-1), a, o, o, o,
              a, o, a^(-1), o, o,
              a^(-1), o, o, a, o,
              a, o, o, o, a^(-1),
              o, o, a^(-1), a, o
              ]);
    t22 := DiagonalJoin(t20,A);
    t14 := DiagonalJoin(t12,-A+1);
    t48 := BlockMatrix(3,3,
            [
            t16,0*t16, 0*t16,
            0*t16, 0*t16, t16,
            0*t16, t16, 0*t16
            ]);
    t80 := DiagonalJoin(<t16,t48,t16^(-1) >);
    t18 := DiagonalJoin(t14, a-2);
    t96 := DiagonalJoin(t80,t16);
    t160 := DiagonalJoin(t80,t80);
    t100 := DiagonalJoin(DiagonalJoin(A,t96),A);
    return GL(d,K)!(eval ("t" cat IntegerToString(d)));
end function;

/*
 * Construct examples which the Meat-Axe cannot handle as follows:
 * Take A normal G, cyclic, non-split extension; take faithful irrep
 * of A and induce up to G. Better let everything be a p-group so that
 * G is really nilpotent!
 */
function DMD(p)
    G<x,g> := Group< x, g | g^p = x^p, x^(p^2) = 1, (x,g) = x^p >;
    G := Codomain(CosetAction(G, sub<G|>));
    A := sub<G | G.1>;
    M := GModule(A,
        [GL(p*(p-1),Rationals())!CompanionMatrix(CyclotomicPolynomial(p^2))]);
    return MatrixGroup(Induction(M,G));
end function;

intrinsic ExampleIrrednil(i::RngIntElt : PaperOnly := false) -> GrpMat
{
 Construct the examples given in the paper. For 1<= i <= 13, return a
 triple <G_i, G_(i,gamma), G_(i,X), G_(i,rho), G_(i,rho,X)>.
}
    requirerange i, 1, 14;
    case i:
    when 1:
        G := RandomifyMatGrp(MatrixGroup<16, Rationals()|
            RegularQuaternionAlgebra(RegularQuaternionAlgebra([1],-1,-1),
                                     -1,-1)>,5,4000);
    when 2:
        G := RandomifyMatGrp(DirectProduct(SylowRatGL(16,2), SylowRatGL(6,3)),
            11, 4000);
        
    when 3:
        list :=
  [[  0,  0,  0, -1,  0,  0,  0,  0,  0,  1, -1, -1,  0,  0,  0,  0, 
  -1,  2, -1, -1,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  
   0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  1, -1, -1,  
   0,  0,  0,  0, -1,  2, -1, -1,  0,  0,  0,  0,  1,  0,  0,  0],
  
   [1, -2,  1,  1,  0,  0,  0,  0,  1, -1,  1,  0,  0,  0,  0,  0,  
   0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  
   0,  0,  0,  0,  1, -2,  1,  1,  0,  0,  0,  0,  1, -1,  1,  0,  
   0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  1,  0],
  
   [0,  1,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  
   1, -1,  1,  0,  0,  0,  0,  0,  1, -1,  0,  1,  0,  0,  0,  0,  
   0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  
   0,  0,  0,  0,  1, -1,  1,  0,  0,  0,  0,  0,  1, -1,  0,  1],
  
   [0,  0,  0,  1,  0,  0,  0,  0,  1, -1,  0,  1,  0,  0,  0,  0,  
   1, -2,  1,  1,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  
   0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  1, -1,  0,  1,  
   0,  0,  0,  0,  1, -2,  1,  1,  0,  0,  0,  0,  1,  0,  0,  0],
  
   [-1,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  
   0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  
   0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  
   0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0, -1],  
  
   [-1,  2, -1, -1,  0,  0,  0,  0,  0,  1, -1, -1,  0,  0,  0,  0,  
   0,  0,  0, -1,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  
   0,  0,  0,  0, -1,  2, -1, -1,  0,  0,  0,  0,  0,  1, -1, -1,  
   0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0, -1,  0]]; 
        G := RandomifyMatGrp(MatrixGroup<8,Rationals()|list>,8,4000)
            ^RandomMat(8:thres:=2,frac);
        
    when 4:
        list :=
            [[ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  2,  3,  2,  1,  0,  2,
            0,  0,  0,  0,  0,  0,  0, -1, -1, -2, -1, -1,  0, -1,
            0,  0,  0,  0,  0,  0,  0,  1,  1,  2,  1,  1,  1,  1,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1, -1,
            0,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  2,  3,  3,  2,  1,  1,
            0,  0,  0,  0,  0,  0,  0, -1, -2, -2, -1, -1, -1, -1],

            [ 1,  1,  1,  1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            -1, -1, -2, -2, -1, -1, -1,  0,  0,  0,  0,  0,  0,  0,
            1,  1,  1,  1,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            -1, -2, -2, -1, -1, -1, -1,  0,  0,  0,  0,  0,  0,  0,
            1,  2,  3,  2,  2,  1,  2,  0,  0,  0,  0,  0,  0,  0,
            -1, -1, -2, -1, -1,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  2,  1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0, -1, -2, -2, -2, -2, -1, -1,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0, -1,
            0,  0,  0,  0,  0,  0,  0,  1,  2,  4,  3,  2,  1,  2,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1, -1, -1,  0,  0,
            0,  0,  0,  0,  0,  0,  0, -1, -2, -2, -1,  0,  0, -1,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1, -2, -2, -1, -1, -1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  1,  2,  2,  1,  1,  1,
            0,  0,  0,  0,  0,  0,  0, -1, -1, -2, -1, -1, -1, -1,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  2,  1,  1,  1,  1,
            0,  0,  0,  0,  0,  0,  0,  1,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  0,  1,
            0,  0,  0,  0,  0,  0,  0, -1, -2, -3, -2, -1,  0, -2,
            0,  0,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1, -1, -1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  1,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1, -1,  0,
            0,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  2,  4,  3,  2,  1,  2,
            0,  0,  0,  0,  0,  0,  0, -1, -3, -4, -3, -2, -1, -2,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1],

            [-1, -2, -2, -2, -2, -1, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0, -1, -1,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  1,  1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0, -1, -1,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            -1, -1, -1, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            1,  2,  3,  2,  1,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            1,  1,  2,  1,  1,  1,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  2,  2,  2,  1,  1,
            0,  0,  0,  0,  0,  0,  0,  1,  2,  2,  1,  0,  0,  1,
            0,  0,  0,  0,  0,  0,  0,  0, -1, -1, -1,  0,  0, -1,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0, -1, -1, -1, -1, -1,  0, -1,
            0,  0,  0,  0,  0,  0,  0, -1, -2, -2, -1, -1, -1, -1],

            [ 0,  0,  0,  0, -1, -1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            1,  3,  4,  3,  2,  1,  2,  0,  0,  0,  0,  0,  0,  0,
            -2, -3, -4, -3, -2, -1, -2,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1],

            [ 0, -1, -2, -1, -1, -1, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            1,  2,  2,  2,  2,  1,  1,  0,  0,  0,  0,  0,  0,  0,
            -2, -3, -4, -3, -2, -1, -2,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1],

            [ 1,  1,  1,  1,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  1,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  1,  1,  1,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0, -1, -2, -1,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1]];
            
        G := RandomifyMatGrp(MatrixGroup<14,Rationals()|list>,8,4000)
               ^RandomMat(14:thres:=2,frac);

    when 5:
        list := [[1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1],

            [ 0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1],

            [-1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1],

            [ 1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1],

            [ 0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,
            0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,
            0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,
            0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0],

            [ 0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,
            0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, -1,  0,  0,
            0,  0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0, -1,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,
            0,  0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0,  0,  0,  0,  0]];
        G := RandomifyMatGrp(MatrixGroup<16,Rationals()| list>,14,4000)^RandomMat(16:thres:=2,frac);
      
    when 6:
        G := RandomifyMatGrp(KronProd(
            RewriteOverRationals(IrrQuatANC(3,5)), SylowRatGL(6,3)),
                                 8, 4000);
    when 7:
        X := SmallGroup(27,3);
        Z := Center(X);
        C := MatrixGroup<2, Rationals() | 
                         CompanionMatrix(CyclotomicPolynomial(3))>;
        f := map< Z -> C | { <Z.0,C.0>, <Z.1,C.1>, <Z.1^2,C.1^2>} >;
        G := RandomifyMatGrp(
            InducedRepMatrixGroup(X,f), 3, 4000)^RandomMat(18:thres:=2);

    when 8:                  
        G :=
            MatrixGroup<12,Rationals()|
            [ 1, 0, 1,-1,-1,-1, 1, 0,-1, 1, 1,-1,
            0, 0, 0, 0, 0,-1, 1, 0, 0, 1, 0,-1,
            0, 0,-1, 1, 1, 0, 0, 0, 1, 0,-1, 0,
            0, 0, 0, 0, 0, 0, 0,-1, 0, 0, 1, 0,
            0,-1,-1, 1, 1, 1,-1, 1, 0,-1,-2, 2,
            0, 0, 0, 0, 1,-1, 0,-1, 1, 1, 0,-1,
            0,-1,-1, 1, 1, 1,-1, 0, 1, 0,-2, 1,
            0, 0, 0,-1, 0, 0, 0, 0, 0, 0, 1, 0,
            0, 1, 0,-1,-1,-1, 1, 0, 0, 1, 1,-1,
            0, 0, 1, 0, 0,-1, 0, 0, 0, 0, 1, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
            0, 0, 0, 1, 0, 0, 0, 1, 0, 0,-1, 1],
 
            [ 0, 0, 0, 0, 0, 0, 0, 1,-1,-1, 0, 1,
            0, 0, 0, 0,-1, 0, 0, 1,-1,-1, 0, 1,
            0, 0, 0, 0, 0, 0,-1, 0, 0, 0, 0, 0,
            1, 0, 1,-1,-1, 0, 0,-1,-1, 0, 2, -1,
            -1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 1,-1, 0, 0,-1, 0, 1,
            0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 1,-1, 0,-1,-1, 1, 1,
            1, 0, 1, 0, 0,-1, 0, 0,-1, 0, 1, 0,
            1,-1,-1, 2, 1, 2,-1, 0, 0,-1,-2, 2,
            1, 0, 1, 0, 0, 0, 0,-1,-1, 0, 1, 0,
            1,-1, 0, 2, 1, 0, 0, 0, 0, 0,-2, 1],
 
            [ 0, 1, 0,-1,-1, 0, 0, 0,-1, 0, 1, 0,
            -1, 1, 0,-1,-1, 0, 0, 0, 0, 0, 1, 0,
            -1, 1, 0, 0, 0,-1, 1, 0, 1, 1, 0,-1,
            0, 0, 1, 0, 0,-1, 0,-1, 0, 0, 1,-1,
            -1, 0,-1, 0, 0, 1, 0, 1, 0,-1,-1, 1,
            -1, 1, 0,-1, 0,-1, 1,-1, 1, 1, 1,-2,
            -1, 0,-1, 0, 1, 0, 0, 0, 1, 0,-1, 0,
            0, 0,-1, 1, 0, 1, 0, 0, 0,-1,-1, 1,
            0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0,-1, 1, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0,-1, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 1, 0,-1, 0, 1],
            
            [-1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0,
            0,-1, 0, 1, 1, 0, 0, 0, 1, 0,-1, 0,
            0, 0, 0, 0, 0,-1, 1, 0, 0, 1, 0,-1,
            0, 0, 0,-1, 0, 0, 0,-1, 0, 0, 1,-1,
            1,-1, 0, 1, 0, 0, 0, 0, 0, 0,-1, 1,
            0, 0, 0, 0, 0,-1, 1, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0,-1, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0,-1, 0, 0, 1, 0,
            -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0,-1, 1, 0, 1, 0, 0, 0,-1,-1, 1,
            0, 0, 0, 0, 0, 0, 0,-1, 0, 0, 0, 0,
            0, 0, 0, 1, 0, 0, 0, 0, 0, 0,-1, 0],
            
            [-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0,-1, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0,-1, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0,-1, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0,-1, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0,-1, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0,-1, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0,-1, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0,-1, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,-1]>;
        M12 := MatrixRing(Rationals(),12);
        rand := GL(96,Rationals())!
            DiagonalJoin( [RandomMat(6:thres:=2) : i in [1..16]])
            * GL(96,Rationals())!IntegralRandomify(GL(96,Rationals())!1,6);
        G := RandomifyMatGrp(KronProd(G,SylowRatGL(8,2)), 10, 4000)^rand;
 
    when 9:
        G := RandomifyMatGrp( RewriteOverRationals(
            IrrDihANC(5,11)), 5, 4000);
    when 10:
        G := RandomifyMatGrp( KronProd(SylowRatGL(16,2),
            SylowRatGL(6,3)), 11, 4000);
    when 11:
        G := RandomifyMatGrp( DMD(5), 5, 4000);
    when 12:
        G := RandomifyMatGrp( RewriteOverRationals(
            RarelyIrrQuatANC(5,11)), 5, 4000);
    when 13:
        G := RandomifyMatGrp($$(2)[1], 100, 4000);
    when 14:
        G := $$(3)[1] ^ RandomMat(8 : thres := 2048);
    end case;

    Q := Rationals();
    Qgamma := NumberField(Polynomial([1,0,-1,1]));
    QX := FunctionField(Rationals());
    Ggamma := ApplyFieldHomMatGrp(G,
        hom<Q -> Qgamma|>)^conjmat(Degree(G),Qgamma);
    GX := ApplyFieldHomMatGrp(G, hom<Q -> QX|>)^conjmat(Degree(G),QX);
    
    if PaperOnly then
        return <G,Ggamma,GX>;
    else
        Qsigma := NumberField(Polynomial([1,2,-3,-3,7,-3,2,5,0,-1,1]));
        Gsigma := ApplyFieldHomMatGrp(G,
            hom<Q -> Qsigma|>)^conjmat(Degree(G),Qsigma);
    
        QsigmaX := FunctionField(Qsigma);
        GsigmaX := ApplyFieldHomMatGrp(Gsigma,
            hom<Qsigma -> QsigmaX|QsigmaX!Qsigma.1>)
            ^conjmat(Degree(G),QsigmaX);
        return <G,Ggamma,GX,Gsigma,GsigmaX>;
    end if;    
end intrinsic;

